{-# LANGUAGE NamedFieldPuns #-}
import Data.List
import E
import ES
import Derivation
import LPO
import Term
import Substitution
import PE
import PES
import TPTP
import UEQParser
import INFParser
import Horn
import Prover
import Util
import System.Environment

-- INF: CoCo infeasibility problem
data Mode = UEQ String | TPTP String | INF String | Help

convertEquation :: PE -> TPTPInput
convertEquation (l, r) = TPTP.CNF "some_equation" "axiom" [(True, Equation l r)]

convertGoal :: PE -> TPTPInput
convertGoal (l, r) = TPTP.CNF "some_equation" "negated_conjecture" [(False, Equation l r)]

convertES :: PES -> TPTP
convertES es = [ convertEquation e | e <- es ]

axiom :: Int -> PE -> E
axiom i pe
  = E { eqn = pe, eqn_id = i, eqn_derivation = Axiom, active = True}

axioms :: PES -> (Int, ES)
axioms pes = (length pes, es)
  where
    es = [ axiom i e | (i, e) <- zip [0..] pes]

parseArgs :: [String] -> Mode
parseArgs ["-h"] = Help
parseArgs ["--help"] = Help
parseArgs [fname] = UEQ fname
parseArgs ["--ueq", fname] = UEQ fname
parseArgs ["--tptp", fname] = TPTP fname
parseArgs ["--inf", fname] = INF fname
parseArgs _ = Help

handleHelp :: IO ()
handleHelp = do
  putStrLn "Usage: toma <options> <file>"
  putStrLn "<file> is in TPTP (UEQ) format, or CoCo INF format."
  putStrLn "Examples: "
  putStrLn "- $ toma inverse_unit.p        # TPTP UEQ format (simple skolemization)"
  putStrLn "- $ toma --ueq  inverse_unit.p # TPTP UEQ format (simple skolemization)"
  putStrLn "- $ toma --tptp inverse_unit.p # TPTP format (split-if transformation)"
  putStrLn "- $ toma --inf  936.trs        # CoCo INF format (split-if transformation)"

printTPTPProof :: PES -> Proof -> IO ()
printTPTPProof pes (Prover.Join { goal, es, deleted }) = do
  putStrLn "% SZS status Unsatisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "The given problem is transformed into a word problem"
  putStrLn "by the split-if encoding (Claessen and Smallbone, 2018)."
  putStrLn "To show the unsatisfiability,"
  putStrLn "it suffices to show that true__ = false__ is valid in the following ES:"
  putStrLn (showPES pes)
  putStrLn "This is an equational proof of true__ = false__:"
  putStrLn (showES (sortById (relevant (eqn_id goal) (goal : es ++ deleted))))
  putStrLn "% SZS output end Proof"
printTPTPProof pes (Prover.Complete { es, prec }) = do
  putStrLn "% SZS status Satisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "The given problem is transformed into a word problem"
  putStrLn "by the split-if encoding (Claessen and Smallbone, 2018)."
  putStrLn "To show the satisfiability,"
  putStrLn "it suffices to show that true__ = false__ is not valid in the following ES:"
  putStrLn (showPES pes)
  putStrLn "The ES admits the following ground-complete ordered rewrite system:"
  putStrLn ("LPO with precedence: " ++ (showPrec prec))
  putStrLn "ES:"
  putStrLn (showOTRS (gtLPO prec) es)
  putStrLn "Since true__ and false__ are not joinable in the ordered rewrite system,"
  putStrLn " true__ = false__ is not valid."
  putStrLn "% SZS output end Proof"
printTPTPProof _ Failure = putStrLn "% SZS status GaveUp : proof failed"

printTPTPTrivial :: IO ()
printTPTPTrivial = do
  putStrLn "% SZS status Satisfiable"
  putStrLn "% SZS output Proof"
  putStrLn "% The given problem admits an trivial model."
  putStrLn "% SZS output end Proof"

-- TODO: suppert 'include' from $TPTP 
handleTPTP :: String -> IO ()
handleTPTP fname = do
  result <- readTPTP fname
  case result of
    Left e -> do
      putStdErr "% SZS status SyntaxError"
      putStdErr (show e)
    Right tptp -> do
      case splitIfTPTP tptp of
        Nothing      -> do
          putStdErr "% SZS status UsageError : The input contains a non-Horn-clause."
        Just (es, goal@(s, t)) -> do
          putStr (showTPTP (convertES es ++ [convertGoal goal]))
          if s == t -- trivial model found
            then printTPTPTrivial
            else do let (next_id, axs) = axioms es
                    p <- prove next_id goal axs
                    printTPTPProof es p

separate :: Int -> [AnnotatedFormula] -> (Int, ES, PES)
separate i [] = (i, [], [])
separate i ((_, _, (Eq, eqn)) : afs) =
  (j, E { eqn, eqn_id = i, eqn_derivation = Axiom, active = True } : es, gs)
  where (j, es, gs) = separate (i + 1) afs
separate i ((_, _, (Neq, g)) : afs) = 
  (j, es, g : gs)
  where (j, es, gs) = separate i afs

skolemize :: [String] -> (Term, Term) -> (Term, Term)
skolemize fs (s, t) =
  (substitute s sigma, substitute t sigma)
  where
    xs = Data.List.union (variables s) (variables t)
    cs = [ c | i <- [1 :: Int ..], let c = "c" ++ show i, notElem c fs ]
    sigma = [ (x, F c []) | (x, c) <- zip xs cs ]

includeDirectories :: IO [String]
includeDirectories = do
  env <- getEnvironment
  case lookup "TPTP" env of
    Nothing  -> return ["."]
    Just dir -> return [".", dir]

printUEQProof :: PE -> PE -> PES -> Proof -> IO ()
printUEQProof _ sko_goal pes (Prover.Join { goal, es, deleted}) = do
  putStrLn "% SZS status Unsatisfiable"
  putStrLn "% SZS output start Proof"
  putStrLn "To show the unsatisfiability,"
  putStrLn ("it suffices to show that " ++ showPE sko_goal   ++ " (skolemized goal) is valid under the axioms.")
  putStrLn "This is an equational proof:"
  putStrLn (showES (sortById (relevant (eqn_id goal) (goal : es ++ deleted))))
  putStrLn "% SZS output end Proof"
printUEQProof orig_goal@(s, t) _ pes (Prover.Complete { es, prec })
  | PE.isGround orig_goal = do
      putStrLn "% SZS status Satisfiable"
      putStrLn "% SZS output start Proof"
      putStrLn "To show the satisfiability,"
      putStrLn ("it suffices to show that " ++ showPE orig_goal ++ " is not valid under the axioms:")
      putStrLn (showPES pes)
      putStrLn "The ES admits the following ground-complete ordered rewrite system:"
      putStrLn ("LPO with precedence: " ++ (showPrec prec))
      putStrLn "ES:"
      putStrLn (showOTRS (gtLPO prec) es)
      putStrLn ("Since " ++ show s ++ " and " ++ show t ++ " are not joinable in the ordered rewrite system,")
      putStrLn "the ground goal is not valid."
      putStrLn "% SZS output end Proof"
  | otherwise = do
      putStrLn "% SZS status GaveUp : goal is not ground"
printUEQProof _ _ _ Failure = putStrLn "% SZS status GaveUp : proof failed"

handleUEQ :: String -> IO ()
handleUEQ filename = do
  dirs <- includeDirectories
  result <- readUEQFile dirs filename
  case result of
    Left e -> do
      putStdErr "% SZS status SyntaxError"
      putStdErr (show e)
    Right afs -> do
      let (n, es, goals) = separate 0 afs
      case goals of
        [goal] -> do
          let fs = functionsInES (goal : toPES es) 
          let skolemized_goal = skolemize fs goal
          p <- prove n skolemized_goal es
          printUEQProof goal skolemized_goal (toPES es) p
        _ ->
          putStdErr "% SZS status UsageError : number of goals must be exactly one"

printINFProof :: PES -> Proof -> IO ()
printINFProof pes (Complete {es, prec} ) = do
  putStrLn "YES"
  putStrLn "Proof: "
  putStrLn "To show the infeasibility of the given CoCo INF problem, "
  putStrLn "we transform the problem into a word problem by the split-if encoding (Claessen and Smallbone, 2018)."
  putStrLn "It suffices to show that the follwing ES does not entail true__ = false__ ."
  putStrLn "ES:"
  putStrLn (showPES pes)
  putStrLn "The ES admits the following equivalent ground-complete ordered rewrite system:"
  putStrLn ("LPO with precedence: " ++ (showPrec prec))
  putStrLn "ES:"
  putStrLn (showOTRS (gtLPO prec) es)
  putStrLn "Since true__ and false__ are not joinable,"
  putStrLn "the original problem is infeasible."
printINFProof _ (Prover.Join {}) = putStrLn "MAYBE"
printINFProof _ Failure = putStrLn "MAYBE"

handleINF :: String -> IO ()
handleINF fname = do
  result <- readINFProblem fname
  case result of
    Left e -> do
      putStdErr "ERROR"
      putStdErr (show e)
    Right inf -> do
      case splitIfINF inf of
        Nothing      -> putStdErr "ERROR"
        Just (es, goal@(s, t)) -> do
          if s == t -- trivial model found
            then do putStrLn "MAYBE" -- TODO: print YES with an horn clauses that admits a trivial model.
                    -- putStrLn "Proof:"
                    -- putStrLn "The given CoCo INF problem is transformed into a satisfiability problem of horn clauses."
                    -- -- TODO: we need to print the horn clauses here...
                    -- putStrLn "Since the horn clauses has a trivial model,"
                    -- putStrLn "the original problem is infeasible."
            else do let (next_id, axs) = axioms es
                    p <- prove next_id goal axs
                    printINFProof es p

dispatch :: Mode -> IO ()
dispatch Help = handleHelp
dispatch (UEQ fname) = handleUEQ fname
dispatch (TPTP fname) = handleTPTP fname
dispatch (INF fname) = handleINF fname

main :: IO ()
main = do
  args <- getArgs
  dispatch (parseArgs args)
