module Prover where

import Data.List
import E
import ES
import Derivation
import PE
import PES
import LPO
import SMT
import Z3
import Term
import LPOEncoding
import Rewriting
import CP
import Join
import KBCompletion
import Substitution
import Horn


findPrec :: PES -> PE -> IO Precedence
findPrec es goal = do
  res <- solve "z3" "toma" z3Input
  case res of
    Nothing -> error "findPrec: no precedence found." -- this error means bug?
    Just res' -> let prec = map fst (sortOn snd [ (fname, -n) | (fname, v) <- res', n <- fromSMTValue2Int v])
                 in return prec
  where
    esWithIds = [("__rulevar" ++ show i, e) | (i, e) <- zip [0..] (es ++ PES.inverse es)]
    -- NOTE: include goal (LPO module requires total precedence)
    functionSymbols = nub [f | (l, r) <- goal : es, f <- (Term.functions l) ++ (Term.functions r)]
    esIds = [ esId | (esId, _) <- esWithIds ]
    -- NOTE: maximize the # of reducibility.
    -- More precisely, maximizes the # of terms t
    -- that satisfies the follwoing condition:
    -- there exists l = r \in E \cup E^{-1} such that
    -- * s encompasses l, and
    -- * l >_lpo r.
    reducibilityOnTerm t =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds, encompass t l ]
    test_terms = nub [ u | (s, t) <- goal : es, u <- [s, t] ]
    maximizedConstraints = [ reducibilityOnTerm t | t <- test_terms ]
    lpoDic = constructDic (map snd esWithIds)
    dicCostraints = andFormula [ impliesFormula (BoolVar name) (encodeLPODic lpoDic s t) | ((s, t), name) <- lpoDic]
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (BoolVar (unsafeLookup (l, r) lpoDic))
                             | (ruleId, (l, r)) <- esWithIds ]
    -- totality is required since precedence is presented as list
    precTotality = Distinct [Var f | f <- functionSymbols]
    z3Input = maxsat functionSymbols
                     (esIds ++ map snd lpoDic)
                     maximizedConstraints
                     (andFormula [orientation, precTotality, dicCostraints])

-- NOTE:
-- Maximizing naively the # of terms t such that t is reducible in (E, >_lpo)
-- seems to generate too many smt variables (?)
-- Encoding itself was also too expensive (?)
-- As a result, the overall procedure became too slow (?)
-- I haven't investigated in detail though...

generateLemmata :: Int -> ES -> Precedence -> (Int, ES)
generateLemmata next_id es prec = (next_id + length ret, ret)
  where cps = selectLemmata [ c | c@((s, t), d) <- cp (gtLPO prec) es,
                                  not (groundJoinable prec es s t),
                                  not (inES (s, t) es) ]
        ret = [ E { eqn = e, eqn_id = i, active = True, eqn_derivation = d}
              | (i, (e, d)) <- zip [next_id..] cps ]

sortByWeight :: [(PE, Derivation)] -> [(PE, Derivation)]
sortByWeight cs = sortBy comp cs
  where comp (e1, _) (e2, _) = compare (PE.size e1) (PE.size e2)

selectLemmata :: [(PE, Derivation)] -> [(PE, Derivation)]
selectLemmata cs = take n (Prover.sortByWeight cs)
  where n = 12

data Proof = Join { goal :: E, es :: ES, deleted :: ES }
           | Complete { es :: ES, prec :: Precedence , deleted :: ES} -- system is ground-complete but goal is not joinable.
           | Failure -- not used for now

prove' :: Int -> PE -> PES -> ES -> ES -> IO Proof
prove' next_id goal@(s, t) initES es deleted_acc = do
  prec <- findPrec (toPES es) goal
  let (next_id', es', deleted) = interreduce next_id prec es
  let next_deleted = deleted ++ deleted_acc
  let gt = gtLPO prec
  let (rw_s, s') = nf es' gt s
  let (rw_t, t') = nf es' gt t
  -- First, the prover checks if s and t are joinable.
  if s' == t'
    then let
           d = Goal { rw_l = rw_s, rw_r = rw_t }
           g' = E { eqn = goal, eqn_id = next_id',
                    active = True, eqn_derivation = d }
          in return (Join { goal = g', es = es', deleted =  next_deleted })
    else if all (\(s'', t'') -> groundJoinable prec es' s'' t'') -- ground completeness checking
                (PES.union (map fst (cp gt es')) initES) -- ES.Union is needed to check equivalence as ESs.
          then return (Complete { es = es', prec = prec, deleted = next_deleted })
          else do let (next_id'', lemmata) = generateLemmata next_id' es' prec
                  let nextES = (es' ++ lemmata)
                  prove' next_id'' goal initES nextES next_deleted

-- goal must be ground 
prove :: Int -> PE -> ES -> IO Proof
prove next_id goal es = prove' next_id goal (toPES es) es []
