module Rewriting (nf) where

import Control.Monad
import Data.Maybe
import E
import ES
import PES
import LPO
import Substitution as S
import Term

type OTRS = (PES, TermOrder)

-- innermost rewriting

data MarkedTerm = NF Term
                | Active String [MarkedTerm]

-- special subsitute for innermost rewriting
substitute :: Term -> Subst -> MarkedTerm
substitute t@(V _) sigma = NF (S.substitute t sigma)
substitute (F f ts) sigma = Active f [ Rewriting.substitute t sigma | t <- ts ]

activate :: Term -> MarkedTerm
activate t@(V _) = NF t
activate (F f ts) = Active f [ activate t | t <- ts ]

rewriteAtRoot :: ES -> TermOrder -> Term -> (Maybe Int, MarkedTerm)
rewriteAtRoot [] gt t = (Nothing, NF t)
rewriteAtRoot (E { eqn = (l, r), eqn_id = i, active = True } : es) gt t
  | Just sig <- match l t,  (S.substitute l sig) `gt` (S.substitute r sig)
  = (Just i, Rewriting.substitute r sig)
  | Just sig <- match r t, (S.substitute r sig) `gt` (S.substitute l sig)
  = (Just i, Rewriting.substitute l sig)
  | otherwise = rewriteAtRoot es gt t
rewriteAtRoot ((E {active = False}) : es) gt t = rewriteAtRoot es gt t

-- NOTE:
-- [Int] is the eqn_ids of E used to nf a term.
-- Almost every nf related function maitains eqn_ids in a way that
-- rules are sorted from (used) later to earlier in innermost rewriting,
-- for efficiency and ease.
-- For example,
-- Let E = { (a): 0 + y -> y, (b): s(x) + y -> s(x+y) }.
-- innermost rewriting from (s(0) + 0) + 0 is:
-- (s(0) + 0) + 0 -b-> s(0+0) + 0 -a-> s(0) + 0 -b-> s(0 + 0) -a-> s(0). 
-- `nf' E > ((s(0) + 0) + 0)`
-- returns [(a) ,(b), (a), (b)]. Note that rewriting with (b) comes last.
-- In contrast, `nf E > ((s(0) + 0) + 0)`
-- returns [(b), (a), (b), (a)]. This should be what users expect.
-- TODO: add more clear example here

nfArgs' :: ES -> TermOrder -> [MarkedTerm] -> [([Int], Term)] -> ([Int], [Term])
-- NOTE: args are parallel. So we don't need to reverse id_acc.
nfArgs' _ _ [] id_term_acc = (Prelude.concat id_acc, reverse term_acc)
  where
    id_acc = map fst id_term_acc
    term_acc = map snd id_term_acc
nfArgs' es gt (t : ts) id_term_acc
  = nfArgs' es gt ts (nf' es gt t : id_term_acc)

nfArgs :: ES -> TermOrder -> [MarkedTerm] -> ([Int], [Term])
nfArgs es gt ts = nfArgs' es gt ts []

nf'' :: ES -> TermOrder -> [Int] -> MarkedTerm -> ([Int], Term)
nf'' _ _ id_acc (NF t) = (id_acc, t)
nf'' es gt id_acc (Active f ts) =
  case rewriteAtRoot es gt (F f nf_args) of
    (Just i, mt)    -> nf'' es gt (i : arg_ids ++ id_acc) mt
    (Nothing, NF t) -> (arg_ids ++ id_acc, t)
    _               -> error "nf"
  where (arg_ids, nf_args) = nfArgs es gt ts

nf' :: ES -> TermOrder -> MarkedTerm -> ([Int], Term)
nf' es gt t = nf'' es gt [] t

-- eqn_ids are sorted from earlier to later in innermost rewriting.
nf :: ES -> TermOrder -> Term -> ([Int], Term)
nf es gt t = (reverse ids, t') -- NOTE: reverse must be done only once here.
  where (ids, t') = nf' es gt (activate t)
