module SMT where

import Data.List

data Exp = Var String
         | Val Int
         | Plus [Exp]
         | Times [Exp]
         deriving Eq

data Formula = And [Formula]
             | Or [Formula]
             | Not Formula
             | Distinct [Exp]
             | Gt Exp Exp
             | Geq Exp Exp
             | Gtn [Exp] -- (> t1 t2 t3 ... tn)
             | Eq Exp Exp
             | BoolVar String
             deriving Eq

instance Show Exp where
    show (Var s)     = s 
    show (Val n)       = show n
    show (Plus []) = "0"
    show (Plus [e]) = show e
    show (Plus es)  = "(+ " ++ intercalate " " [show e | e <- es] ++ ")"
    show (Times []) = "1"
    show (Times [e]) = show e
    show (Times es)  = "(* " ++ intercalate " " [show e | e <- es] ++ ")"

instance Show Formula where
    show (BoolVar s) = s
    show (And []) = "true"
    show (And [f]) = show f
    show (And fs) = "(and " ++ intercalate " " [show f | f <- fs] ++ ")"
    show (Or []) = "false"
    show (Or [f]) = show f
    show (Or fs) = "(or " ++ intercalate " " [show f | f <- fs] ++ ")"
    show (Not f) = "(not " ++ show f ++ ")"
    show (Distinct es) =  "(distinct " ++ intercalate " " [show e | e <- es] ++ ")"
    show (Gt e1 e2) = "(> " ++ show e1 ++ " " ++ show e2 ++ ")"
    show (Geq e1 e2) = "(>= " ++ show e1 ++ " " ++ show e2 ++ ")"
    show (Gtn es) =  "(> " ++ intercalate " " [show e | e <- es] ++ ")"
    show (Eq e1 e2) = "(= " ++ show e1 ++ " " ++ show e2 ++ ")"

data Command =
     DeclareInt String
   | DeclareBool String
   | Assert Formula
   | AssertSoft Formula
   | CheckSat
   | GetValue [String]

type SMTInput = [Command]

data SMTValue = SMTBoolValue Bool | SMTIntValue Int
                deriving Show

type SMTOutput = Maybe [(String, SMTValue)]

instance Show Command where
    show (DeclareInt varname) = "(declare-const "  ++ varname ++ " Int)"
    show (DeclareBool varname) = "(declare-const "  ++ varname ++ " Bool)"
    show (Assert f) = "(assert " ++ show f ++ ")"
    show (AssertSoft f) = "(assert-soft " ++ show f ++ ")"
    show CheckSat = "(check-sat)"
    show (GetValue vars) = "(get-value (" ++ intercalate " " [ v | v <- vars ] ++ "))"

-- smart constructors

andFormula :: [Formula] -> Formula
andFormula [f] = f -- propagete partial evaluation
andFormula fs
  | falseFormula `elem` fs' = falseFormula
  | otherwise = case fs' of
                  [] -> trueFormula
                  _  -> And fs'
  where fs' = [f | f <- fs, f /= trueFormula]

orFormula :: [Formula] -> Formula
orFormula [f] = f -- propagate partial evaluation
orFormula fs
  | trueFormula `elem` fs' = trueFormula
  | otherwise = case fs' of
                  [] -> falseFormula
                  _  -> Or fs'
  where fs' = [f | f <- fs, f /= falseFormula]

notFormula :: Formula -> Formula
notFormula f
  | f == trueFormula = falseFormula
  | f == falseFormula = trueFormula
  | otherwise = Not f

impliesFormula :: Formula -> Formula -> Formula
impliesFormula p q = orFormula [notFormula p, q]

trueFormula :: Formula
trueFormula = And []

falseFormula :: Formula
falseFormula = Or []

isSMTBoolTrue :: SMTValue -> Bool
isSMTBoolTrue (SMTBoolValue True) = True
isSMTBoolTrue _ = False

fromSMTValue2Int :: SMTValue -> [Int]
fromSMTValue2Int (SMTIntValue n) = [n]
fromSMTValue2Int _ = []

smtValueUnsafeCastInt :: SMTValue -> Int
smtValueUnsafeCastInt (SMTIntValue i) = i
-- TODO: this behaviour seems confusing...
smtValueUnsafeCastInt (SMTBoolValue True) = 1
smtValueUnsafeCastInt (SMTBoolValue False) = 0

showSMTInput :: SMTInput -> String
showSMTInput commands = intercalate "\n" [show c | c <- commands]

generateSMTInput :: [String] -> [String] -> Formula -> SMTInput
generateSMTInput ivs bvs formula = commands
  where commands =  [DeclareInt v | v <- ivs]
                 ++ [DeclareBool v | v <- bvs]
                 ++ [Assert formula, CheckSat, GetValue ivs ]

maxsat :: [String] -> [String] -> [Formula] -> Formula -> SMTInput
maxsat ivs bvs ss h = commands
  where commands =  [DeclareInt v | v <- ivs]
                 ++ [DeclareBool v | v <- bvs]
                 ++ [AssertSoft s | s <- ss]
                 ++ [Assert h, CheckSat, GetValue (ivs ++ bvs)]

-- renaming

rename :: String -> [(String, String)] -> String
rename x dic =
  case lookup x dic of
    Just y -> y
    Nothing -> error "rename: not found"

variables :: SMTInput -> [String]
variables [] = []
variables ((DeclareInt v) : cs) = v : variables cs
variables ((DeclareBool v) : cs) = v : variables cs
variables (_ : cs) = variables cs

renameExp :: Exp -> [(String, String)] -> Exp
renameExp (Var x) dic = Var (rename x dic)
renameExp v@(Val _) dic =  v
renameExp (Plus es) dic = Plus [ renameExp e dic | e <- es]
renameExp (Times es) dic = Times [ renameExp e dic | e <- es]

renameFormula :: Formula -> [(String, String)] -> Formula
renameFormula (And fs) dic = And [ renameFormula f dic | f <- fs ]
renameFormula (Or fs) dic = Or [ renameFormula f dic | f <- fs ]
renameFormula (Not f) dic = Not (renameFormula f dic)
renameFormula (Distinct es) dic = Distinct [ renameExp e dic | e <- es]
renameFormula (Gt e1 e2) dic = Gt (renameExp e1 dic) (renameExp e2 dic)
renameFormula (Geq e1 e2) dic = Geq (renameExp e1 dic) (renameExp e2 dic)
renameFormula (Gtn es) dic = Gtn [ renameExp e dic | e <- es ]
renameFormula (Eq e1 e2) dic = Eq (renameExp e1 dic) (renameExp e2 dic)
renameFormula (BoolVar x) dic = BoolVar (rename x dic)

renameSMTInput :: SMTInput -> [(String, String)] -> SMTInput
renameSMTInput [] _ = []
renameSMTInput ((DeclareInt x) : cs) dic
  = (DeclareInt (rename x dic)) : (renameSMTInput cs dic)
renameSMTInput ((DeclareBool x) : cs) dic
  = (DeclareBool (rename x dic)) : (renameSMTInput cs dic)
renameSMTInput ((Assert f) : cs) dic
  = (Assert (renameFormula f dic)) : (renameSMTInput cs dic)
renameSMTInput ((AssertSoft f) : cs) dic
  = (AssertSoft (renameFormula f dic)) : (renameSMTInput cs dic)
renameSMTInput (CheckSat : cs) dic = CheckSat : (renameSMTInput cs dic)
renameSMTInput (GetValue xs : cs) dic
  = (GetValue [ rename x dic | x <- xs ]) : (renameSMTInput cs dic)
